name = "echofilter"
path = name.lower().replace("-", "_").replace(" ", "_")
version = "1.1.0"
author = "Scott C. Lowe"
author_email = "scott.code.lowe@gmail.com"
description = "Remove echosounder noise by identifying the ocean floor and entrained air at the ocean surface."
url = "https://github.com/DeepSenseCA/echofilter"
license = "AGPLv3"
